/*

  stacker.ino - software for an Arduino-based focus stacking and camera controller
  Written by Rylee Isitt
  Rev. 23 Feb 2014

  This softare is freeware, although I will not turn away donations ;)

*/

#include <EEPROM.h>
#include <avr/pgmspace.h>
#include "nokiaLCD.h"
#include "shiftRegister8.h"
#include "stepper.h"
#include "camera.h"
#include "MemoryFree.h"

// ----------------------------------------------------------------------------------------------------
// The values of the variables below can be modified to configure the software for different hardware
// ----------------------------------------------------------------------------------------------------

// pinout configuration
// recommended pinout for the Arduino UNO
#define PIN_IR 2
#define PIN_DATA 3
#define PIN_LATCH 4
#define PIN_CLOCK 5
#define PIN_LCD_SCE   8
#define PIN_LCD_RESET 9
#define PIN_LCD_DC    10
#define PIN_LCD_SDIN  11
#define PIN_LCD_SCLK  12
#define PIN_LCD_LED 13

// motor configuration defaults (see stepper class for more)
// many of these parameters can be adjusted when the program is running, and new values will be saved in EEPROM
boolean relaxMotor = true; // When true, motor brakes between when not in use - this reduces holding torque but can prevent overheating and is the default/recommended setting. Setting this to false will enable full holding torque but may cause a lot of heat
unsigned int umPerRev = 200; // to figure out this number, advance the platform through many revolutions, measure the distance moved with calipers or a micrometer, divide this distance by the number of revolutions, and round to the nearest whole number.

// camera configuration defaults (see camera class for more)
// many of these parameters can be adjusted when the program is running, and new values will be saved in EEPROM
unsigned int exposureTime = 100; // time (in milliseconds) that shutter release relay is kept closed
boolean cableTrigger = true; // determines if a camera is triggered over the shutter release cable

// general remote configuration
#define REMOTE_RELEASEDELAY 100000L // elapsed time (in microseconds) after which if no bits have been received we assume that no button is being pressed
#define BT_NONE 0 // not a real button - a placeholder code for when no buttons are being pressed

// specific remote configuration
// This project has built-in support for one of the following remotes, simply uncomment the applicable code and comment out the rest
// If using a different remote, use the irdecode sketch (http://ryleeisitt.ca/articles/building-a-focus-stacking-controller/#software) to determine the values to use
// Make sure any remote you choose has enough buttons. I recommend remotes for CD players or TVs
// You can also emulate one of the remotes below using programmable remotes

// Kenwood RC-P400, no longer as easy to find but a very convenient size, for a DPC-731 or DPC-531 portable CD player
// Lacks directional buttons and an enter button, but can use other buttons to do the job
// Not in the Harmony DB
#define REMOTE_MINBITS 32 // the number of bits that must be received before a code can be considered received
#define REMOTE_MAXBITS 32 // the number of bits received after which a code must be considered received
#define REMOTE_HIGHMAX 2700UL // elapsed time (in microseconds) since last down-sloping IR signal edge above which the code reading will reset
#define REMOTE_HIGHMIN 1600UL // elapsed time (in microseconds) since last down-sloping IR signal edge that acts as a theshhold between a 0 and 1 bit
#define IR_TRIGGER FALLING
#define BT_1 1838317695 // 1
#define BT_2 1838301375 // 2
#define BT_3 1838334015 // 3
#define BT_4 1838293215 // 4
#define BT_5 1838325855 // 5
#define BT_6 1838309535 // 6
#define BT_7 1838342175 // 7
#define BT_8 1838289135 // 8
#define BT_9 1838321775 // 9
#define BT_0 1838285055 // 0
#define BT_ENTER 1838305455 // +10 on Kenwood remote
#define BT_REWIND 1838311575 // Rewind on Kenwood remote
#define BT_FASTFORWARD 1838344215 // Fast Forward on Kenwood remote
#define BT_UP 1838322285 // Volume Up on Kenwood remote
#define BT_DOWN 1838289645 // Volume Down on Kenwood remote
#define BT_LEFT 1838289645 // Volume Down on Kenwood remote
#define BT_RIGHT 1838322285 // Volume Up on Kenwood remote
#define BT_STOP 1838319735 // Stop on Kenwood remote
#define BT_START 1838336055 // Play/pause on Kenwood remote
#define BT_MENU 1838348295 // P. Mode on Kenwood remote
#define BT_HALFPRESS 1838285820 // AI AUTO on Kenwood remote
#define BT_FULLPRESS 1838302140 // AI TIMER on Kenwood remote
#define BT_BACKLIGHT 1838318460 // PRESENCE on Kenwood remote

// AA59-00666A Remote for Samsung UN39EH5003F LCD TV
// In the Harmony DB, so you can emulate this remote with a Logitech Harmony device
// #define REMOTE_MINBITS 32 // the number of bits that must be received before a code can be considered received
// #define REMOTE_MAXBITS 32 // the number of bits received after which a code must be considered received
// #define REMOTE_HIGHMAX 2700UL // elapsed time (in microseconds) since last down-sloping IR signal edge above which the code reading will reset
// #define REMOTE_HIGHMIN 1600UL // elapsed time (in microseconds) since last down-sloping IR signal edge that acts as a theshhold between a 0 and 1 bit
// #define IR_TRIGGER FALLING
// #define BT_1 3772784863 // 1
// #define BT_2 3772817503 // 2
// #define BT_3 3772801183 // 3
// #define BT_4 3772780783 // 4
// #define BT_5 3772813423 // 5
// #define BT_6 3772797103 // 6
// #define BT_7 3772788943 // 7
// #define BT_8 3772821583 // 8
// #define BT_9 3772805263 // 9
// #define BT_0 3772811383 // 0
// #define BT_ENTER 3772782313 // Enter
// #define BT_REWIND 3772818013 // Rewind
// #define BT_FASTFORWARD 3772781293 // Fast Forward
// #define BT_UP 3772778233 // Up
// #define BT_DOWN 3772810873 // Down
// #define BT_LEFT 3772819033 // Left
// #define BT_RIGHT 3772794553 // Right
// #define BT_STOP 3772801693 // Stop
// #define BT_START 3772834333 // Play
// #define BT_MENU 3772799143 // MENU
// #define BT_HALFPRESS 3772831273 // CH LIST
// #define BT_FULLPRESS 3772837903 // MUTE
// #define BT_BACKLIGHT 3772793023 // Power

// Sony Bravia KDL-55HX850 TV Protocol, will work for many different Sony TV remotes
// This is in the Harmony DB, so you can emulate this remote with a Logitech Harmony device
// #define REMOTE_MINBITS 12 // the number of bits that must be received before a code can be considered received
// #define REMOTE_MAXBITS 15 // the number of bits received after which a code must be considered received
// #define REMOTE_HIGHMAX 2700UL // elapsed time (in microseconds) since last down-sloping IR signal edge above which the code reading will reset
// #define REMOTE_HIGHMIN 1600UL // elapsed time (in microseconds) since last down-sloping IR signal edge that acts as a theshhold between a 0 and 1 bit
// #define IR_TRIGGER RISING
// #define BT_1 16 // 1
// #define BT_2 2064 // 2
// #define BT_3 1040 // 3
// #define BT_4 3088 // 4
// #define BT_5 528 // 5
// #define BT_6 2576 // 6
// #define BT_7 1552 // 7
// #define BT_8 3600 // 8
// #define BT_9 272 // 9
// #define BT_0 2320 // 0
// #define BT_ENTER 2672 // Enter
// #define BT_REWIND 27881 // Rewind (15 bits)
// #define BT_FASTFORWARD 7401 // Fast Forward (15 bits)
// #define BT_UP 752 // Up
// #define BT_DOWN 2800 // Down
// #define BT_LEFT 720 // Left
// #define BT_RIGHT 3280 // Right
// #define BT_STOP 3305 // Stop (15 bits)
// #define BT_START 11497 // Play (15 bits)
// #define BT_MENU 112 // MENU
// #define BT_HALFPRESS 2192 // Channel down
// #define BT_FULLPRESS 144 // Channel up
// #define BT_BACKLIGHT 1168 // Volume up

// shift register bitmasks for camera reed relays
const byte cameraRelay1 = B01000000; // shift register output pin QG/Q6
const byte cameraRelay2 = B10000000; // shift register output pin QH/Q7

// shift register bitmasks for stepper motor
const byte motorEnableA = B00000001; // shift register output pin QA/Q0
const byte motorEnableB = B00000010; // shift register output pin QB/Q1
const byte motorA1 = B00000100; // shift register output pin QC/Q2
const byte motorA2 = B00001000; // shift register output pin QD/Q3
const byte motorB1 = B00010000; // shift register output pin QE/Q4
const byte motorB2 = B00100000; // shift register output pin QF/Q5

// instantiate LCD
nokiaLCD lcd;

// instantiate shift register
shiftRegister8 shiftRegister;

// instantiate stepper motor
stepper stepperMotor(&shiftRegister, motorEnableA, motorEnableB, motorA1, motorA2, motorB1, motorB2);

// instantiate camera
camera cam(&shiftRegister, cameraRelay1, cameraRelay2);

// variables for getting IR codes
// if using a remote that uses 16 bits or less for its code
// change the variable type of code and lastCode to unsigned int
volatile unsigned long lastCode = BT_NONE;
volatile unsigned long lastBitTimestamp = 0;

// variables for tracking saved stepper positions
long startPos = 0; // measured in half steps, even if using full step mode
long endPos = 0; // measured in half steps, even if using full step mode

// ----------------------------------------------------------------------------------------------------
// This function loads config values from EEPROM
// ----------------------------------------------------------------------------------------------------

void loadConfig()
{

    // read first 2 bytes of EEPROM - if they are not set to 64 and 128 respectively, assume that the defaults have never been written in EEPROM
    if ((EEPROM.read(0) == 64) && (EEPROM.read(1) == 128))   // EEPROM values have been set
    {
        // load values from EEPROM
        stepperMotor.halfMode = (EEPROM.read(2) > 0);
        stepperMotor.reverseMotor = (EEPROM.read(3) > 0);
        stepperMotor.stepsPerRev = ((unsigned int)EEPROM.read(5) << 8) + EEPROM.read(4);
        umPerRev = ((unsigned int)EEPROM.read(7) << 8) + EEPROM.read(6);
        stepperMotor.maxRPM = ((unsigned int)EEPROM.read(9) << 8) + EEPROM.read(8);
        relaxMotor = (EEPROM.read(10) > 0);
        if (EEPROM.read(11) == 64) // check for this config revision
        {
            stepperMotor.accelerate = (EEPROM.read(12) > 0);
            exposureTime = ((unsigned int)EEPROM.read(14) << 8) + EEPROM.read(13);
            cableTrigger = (EEPROM.read(15) > 0);
            cam.reverseRelays = (EEPROM.read(16) > 0);
        }
    }
    else     // EEPROM values have never been set
    {
        // write current defaults to EEPROM
        EEPROM.write(0, 64);
        EEPROM.write(1, 128);
        saveConfig();
    }

}

// ----------------------------------------------------------------------------------------------------
// This function saves config values to EEPROM
// ----------------------------------------------------------------------------------------------------

void saveConfig()
{
    const unsigned int firstByteMask = 255;
    const unsigned int secondByteMask = 65280;
    EEPROM.write(2, stepperMotor.halfMode);
    EEPROM.write(3, stepperMotor.reverseMotor);
    EEPROM.write(4, stepperMotor.stepsPerRev & firstByteMask); // first (rightmost) byte
    EEPROM.write(5, (stepperMotor.stepsPerRev & secondByteMask) >> 8); // second (leftmost) byte
    EEPROM.write(6, umPerRev & firstByteMask); // first (rightmost) byte
    EEPROM.write(7, (umPerRev & secondByteMask) >> 8); // second (leftmost) byte
    EEPROM.write(8, stepperMotor.maxRPM & firstByteMask); // first (rightmost) byte
    EEPROM.write(9, (stepperMotor.maxRPM & secondByteMask) >> 8); // second (leftmost) byte
    EEPROM.write(10, relaxMotor);
    EEPROM.write(11, 64); // flag for config revision
    EEPROM.write(12, stepperMotor.accelerate);
    EEPROM.write(13, exposureTime & firstByteMask); // first (rightmost) byte
    EEPROM.write(14, (exposureTime & secondByteMask) >> 8); // second (leftmost) byte
    EEPROM.write(15, cableTrigger);
    EEPROM.write(16, cam.reverseRelays);
}

// ----------------------------------------------------------------------------------------------------
// This function clears the EEPROM
// ----------------------------------------------------------------------------------------------------

void clearEEPROM()
{
    for (unsigned int i = 0; i <= 16; i++)
    {
        EEPROM.write(i, 255);
    }
    lcd.cls();
    lcd.writeLine_P(PSTR("EEPROM cleared"));
    delay(2000);
}

// ----------------------------------------------------------------------------------------------------
// This function is called by an interrupt when the IR receiver breakout board signal falls
// It measures the intervals between falling edges and uses these intervals to decode the IR signal
// ----------------------------------------------------------------------------------------------------

void irFire()
{
    /*
      Deciphering codes from the remote works by watching the elapsed time between down-sloping or up-sloping edges (as a high signal transitions to a low signal or vice versa)
      When this happens, irFire() is triggered by an interrupt
      ___     ____
         |___|    |____
         a        b
      In the waveform above, this elapsed time is the period between "a" and "b" if interrupts are triggered on FALLING
      If the elapsed time is below REMOTE_HIGHMIN, this is interpreted as a 0
      If the elapsed time is between REMOTE_HIGHMIN and REMOTE_HIGHMAX, this is interpreted as a 1
      If the elapsed time is above REMOTE_HIGHMAX, this resets code reading and captures the current code if it's between REMOTE_MINBITS and REMOTE_MAXBITS long
      Once REMOTE_MAXBITS bits have been received, code reading is reset and the code is captured
    */

    static unsigned long code = 0;
    static byte bitCount = 0;
    static unsigned long curTime = 0;
    curTime = micros();

    // watch the bits coming from the receiver
    if ((curTime - lastBitTimestamp) < REMOTE_HIGHMIN)   // zero
    {
        bitCount++;
        code <<= 1;
    }
    else if ((curTime - lastBitTimestamp) < REMOTE_HIGHMAX)     // one
    {
        bitCount++;
        code <<= 1;
        code++;
    }
    else     // begin a new code
    {
        if ((bitCount >= REMOTE_MINBITS) && (bitCount <= REMOTE_MAXBITS)) lastCode = code;
        code = 0;
        bitCount = 0;
    }

    // REMOTE_MAXBITS is reached, so capture code and begin anew
    if (bitCount == REMOTE_MAXBITS)
    {
        lastCode = code;
        bitCount = 0;
        code = 0;
    }

    lastBitTimestamp = curTime;

}

// ----------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------

boolean pressState = false;
void setPressState()
{
    if (!pressState) lastCode = BT_NONE; // a button was not recently pressed so clear lastCode
	// because lastBitTimestamp is volatile, it may be updated in the below statement after micros() is fetched
	// so need to used signed math here, and make sure that REMOTE_RELEASEDELAY is a signed long
    pressState = ((long)(micros() - lastBitTimestamp) < REMOTE_RELEASEDELAY); // check if a button up event has occurred
}

// ----------------------------------------------------------------------------------------------------
// This function returns a byte corresponding to the digit pressed on the remote
// 255 is returned if no digit button has been pressed
// ----------------------------------------------------------------------------------------------------

byte lastRemoteDigit()
{
    byte remoteDigit = 255;
    if (!pressState)
    {
        if (lastCode == BT_1) remoteDigit = 1;
        else if (lastCode == BT_2) remoteDigit = 2;
        else if (lastCode == BT_3) remoteDigit = 3;
        else if (lastCode == BT_4) remoteDigit = 4;
        else if (lastCode == BT_5) remoteDigit = 5;
        else if (lastCode == BT_6) remoteDigit = 6;
        else if (lastCode == BT_7) remoteDigit = 7;
        else if (lastCode == BT_8) remoteDigit = 8;
        else if (lastCode == BT_9) remoteDigit = 9;
        else if (lastCode == BT_0) remoteDigit = 0;
    }
    return remoteDigit;
}

// ----------------------------------------------------------------------------------------------------
// This function adds digits to the end (rightmost) of an unsigned integer, and prevents overflow
// ----------------------------------------------------------------------------------------------------

boolean addDigitUI(unsigned int &param, byte digit, boolean appendMode)
{
    if (appendMode)
    {
        if ((param <= 6552) || ((param == 6553) && (digit <= 5)))
        {
            param = (param * 10) + digit;
        }
        else
        {
            param = 65536;
        }
    }
    else
    {
        param = digit;
    }
    return true;
}

// ----------------------------------------------------------------------------------------------------
// This function draws a scrollbar on the right side
// ----------------------------------------------------------------------------------------------------

void vertScrollbar(byte curPos, byte lastPos)
{
    lcd.writeXY(lcd.cols-1, 1, "(", false, true);
    lcd.writeXY(lcd.cols-1, lcd.rows-1, ")", false, true);
    for (int y=2; y<lcd.rows-1; y++)
        lcd.writeXY(lcd.cols-1, y, "*", false, true);
    byte index = ((curPos * 11) / lastPos) + 1; // ranges from 1 to 12
    byte row = ((index - 1) / 4) + 2; // ranges from 2 to 4
    char quarter = index - (row-2)*4 + 42;
    lcd.writeXY(lcd.cols-1, row, String(quarter), false, true);
}

// ----------------------------------------------------------------------------------------------------
// This function handles the configuration for the stepper motor in the stacking mode
// ----------------------------------------------------------------------------------------------------

void motorConfig()
{

    boolean updateDisplay = true;
    boolean saveNeeded = false;
    byte curParam = 0;
    byte lastParam = 6;
    boolean appendMode = false;
    boolean inMenu = true;

    while (inMenu)
    {
        if (updateDisplay)     // update display
        {
            // figure out which parameters to display
            byte firstToDisplay = 0;
            if (curParam >= (lcd.rows - 1)) firstToDisplay = curParam - lcd.rows + 2;
            byte lastToDisplay = firstToDisplay + lcd.rows - 2;

            lcd.cls();
            lcd.write_P(PSTR(" MOTOR       "),true);
            if (!cam.halfPressState) lcd.writeLine_P(PSTR(" "),true);
            else lcd.writeLine_P(PSTR("3"),true,true);
            if ((firstToDisplay <= 0) && (lastToDisplay >= 0))
            {
                lcd.write_P(PSTR("Half-step:"),curParam==0);
                lcd.writeLine(String(stepperMotor.halfMode),curParam==0,true);
            }
            if ((firstToDisplay <= 1) && (lastToDisplay >= 1))
            {
                lcd.write_P(PSTR("Reverse:"),curParam==1);
                lcd.writeLine(String(stepperMotor.reverseMotor),curParam==1,true);
            }
            if ((firstToDisplay <= 2) && (lastToDisplay >= 2))
            {
                lcd.write_P(PSTR("St/rev:"),curParam==2);
                lcd.writeLine(String(stepperMotor.stepsPerRev),curParam==2);
            }
            if ((firstToDisplay <= 3) && (lastToDisplay >= 3))
            {
                lcd.write_P(PSTR("'"),curParam==3,true);
                lcd.write_P(PSTR("m/rev:"),curParam==3);
                lcd.writeLine(String(umPerRev),curParam==3);
            }
            if ((firstToDisplay <= 4) && (lastToDisplay >= 4))
            {
                lcd.write_P(PSTR("Max RPM:"),curParam==4);
                lcd.writeLine(String(stepperMotor.maxRPM),curParam==4);
            }
            if ((firstToDisplay <= 5) && (lastToDisplay >= 5))
            {
                lcd.write_P(PSTR("Relax motor:"),curParam==5);
                lcd.writeLine(String(relaxMotor),curParam==5,true);
            }
            if ((firstToDisplay <= 6) && (lastToDisplay >= 6))
            {
                lcd.write_P(PSTR("Accelerate:"),curParam==6);
                lcd.writeLine(String(stepperMotor.accelerate),curParam==6,true);
            }
            vertScrollbar(curParam, lastParam); // draw scroll bar on right side
            updateDisplay = false;
        }

        setPressState();
        byte remoteDigit = lastRemoteDigit();
        if (!pressState && (lastCode != BT_NONE))   // a button has been released
        {
            if (remoteDigit != 255) // a digit was pressed
            {
                if (curParam==0) stepperMotor.halfMode = (remoteDigit > 0);
                if (curParam==1) stepperMotor.reverseMotor = (remoteDigit > 0);
                if (curParam==2) appendMode = addDigitUI(stepperMotor.stepsPerRev,remoteDigit,appendMode);
                if (curParam==3) appendMode = addDigitUI(umPerRev,remoteDigit,appendMode);
                if (curParam==4) appendMode = addDigitUI(stepperMotor.maxRPM,remoteDigit,appendMode);
                if (curParam==5) relaxMotor = (remoteDigit > 0);
                if (curParam==6) stepperMotor.accelerate = (remoteDigit > 0);
                saveNeeded = true;
            }
            else
            {
                // enforce minimum values
                if (stepperMotor.stepsPerRev < 4) stepperMotor.stepsPerRev = 4; // this is 90 degrees per step
                if (umPerRev < 1) umPerRev = 1;
                if (stepperMotor.maxRPM < 1) stepperMotor.maxRPM = 1;

                // respond to button presses
                if (lastCode == BT_BACKLIGHT) lcd.toggleBacklight();
                else if (lastCode == BT_HALFPRESS) cam.setHalfPress(!cam.halfPressState);
                else if (lastCode == BT_FULLPRESS) cam.setFullPress(false);
                else if (lastCode == BT_UP)
                {
                    if (curParam-- == 0) curParam = lastParam; // cycle selected field up
                }
                else if (lastCode == BT_DOWN)
                {
                    if (++curParam > lastParam) curParam = 0; // cycle selected field down
                }
                else if (lastCode == BT_ENTER)   // toggle booleans
                {
                    if (curParam==0) stepperMotor.halfMode = !stepperMotor.halfMode;
                    if (curParam==1) stepperMotor.reverseMotor = !stepperMotor.reverseMotor;
                    if (curParam==5) relaxMotor = !relaxMotor;
                    if (curParam==6) stepperMotor.accelerate = !stepperMotor.accelerate;
                    saveNeeded = true;
                }
                else if (lastCode == BT_MENU) inMenu = false;
                appendMode = false;
            }
            updateDisplay = true;
        }
        else if (pressState)     // a button is being pressed
        {
            if (lastCode == BT_FULLPRESS) cam.setFullPress(true);
        }
    }

    // save configuration
    if (saveNeeded) saveConfig();

}

// ----------------------------------------------------------------------------------------------------
// This function handles the configuration for the camera in the stacking mode
// ----------------------------------------------------------------------------------------------------

void cameraConfig()
{

    boolean updateDisplay = true;
    boolean saveNeeded = false;
    byte curParam = 0;
    byte lastParam = 2;
    boolean appendMode = false;
    boolean inMenu = true;

    while (inMenu)
    {
        if (updateDisplay)     // update display
        {
            lcd.cls();
            lcd.write_P(PSTR(" CAMERA      "),true);
            if (!cam.halfPressState) lcd.writeLine_P(PSTR(" "),true);
            else lcd.writeLine_P(PSTR("3"),true,true);
            lcd.write_P(PSTR("Trigger:"),curParam==0);
            lcd.writeLine(String(cableTrigger),curParam==0,true);
            lcd.write_P(PSTR("Swap Relays:"),curParam==1);
            lcd.writeLine(String(cam.reverseRelays),curParam==1,true);
            lcd.write_P(PSTR("TV:"),curParam==2);
            lcd.write(String(exposureTime),curParam==2);
            lcd.writeLine_P(PSTR("ms"),curParam==2);
            updateDisplay = false;
        }

        setPressState();
        byte remoteDigit = lastRemoteDigit();
        if (!pressState && (lastCode != BT_NONE))   // a button has been released
        {
            if (remoteDigit != 255) // a digit was pressed
            {
                if (curParam==0) cableTrigger = (remoteDigit > 0);
                if (curParam==1) cam.reverseRelays = (remoteDigit > 0);
                if (curParam==2) appendMode = addDigitUI(exposureTime,remoteDigit,appendMode);
                saveNeeded = true;
            }
            else
            {
                // enforce minimum values
                if (exposureTime < 1) exposureTime = 1;

                // respond to button presses
                if (lastCode == BT_BACKLIGHT) lcd.toggleBacklight();
                else if (lastCode == BT_HALFPRESS) cam.setHalfPress(!cam.halfPressState);
                else if (lastCode == BT_FULLPRESS) cam.setFullPress(false);
                else if (lastCode == BT_UP)
                {
                    if (curParam-- == 0) curParam = lastParam; // cycle selected field up
                }
                else if (lastCode == BT_DOWN)
                {
                    if (++curParam > lastParam) curParam = 0; // cycle selected field down
                }
                else if (lastCode == BT_ENTER)
                {
                    if (curParam==0) cableTrigger = !cableTrigger;
                    if (curParam==1) cam.reverseRelays = !cam.reverseRelays;
                    saveNeeded = true;
                }
                else if (lastCode == BT_MENU) inMenu = false;
                appendMode = false;
            }
            updateDisplay = true;
        }
        else if (pressState)     // a button is being pressed
        {
            if (lastCode == BT_FULLPRESS) cam.setFullPress(true);
        }
    }

    // save configuration
    if (saveNeeded) saveConfig();

}

// ----------------------------------------------------------------------------------------------------
// This function performs the stacking
// ----------------------------------------------------------------------------------------------------

boolean doStack(unsigned int slices, unsigned int interval, unsigned int bracket)
{

    // timing
    unsigned long delayStartTime = 0;

    // check if exposure lock/half press mode is on at the start
    // if so, we will not automatically engage/disengage the half-press
    boolean autoHalfPress = !cam.halfPressState;

    // initial screen
    lcd.cls();
    lcd.write_P(PSTR(" STACKING    "),true);
    if (!cam.halfPressState)
    {
        lcd.writeLine_P(PSTR(" "),true);
    }
    else
    {
        lcd.writeLine_P(PSTR("3"),true,true);
    }
    lcd.write_P(PSTR("Slice 0/"));
    lcd.writeLine(String(slices));
    lcd.write_P(PSTR("Shot 0/"));
    lcd.writeLine(String(slices*bracket));
    lcd.writeLine_P(PSTR("!\"\"\"\"\"\"\"\"\"\"\"\"#"),false,true);
    lcd.write_P(PSTR("5"),false,true);
    lcd.writeLine_P(PSTR(":Stop"));

    // stack
    unsigned int shot = 0;
    for (int i=0; i<slices; i++)
    {
        unsigned int waitQuarterTime = (interval - exposureTime) / 4;

        // calculate target position
        long gotoPos = startPos + ((long(i)*(endPos-startPos)) / (long(slices) - 1L));

        // go to target position
        stepperMotor.seek(gotoPos);
        do
        {
            stepperMotor.update();
            if (lastCode == BT_STOP)
            {
                stepperMotor.stop();
                return false;
            }
        }
        while (stepperMotor.isMoving);
        if (relaxMotor) stepperMotor.brake();

        // take pictures
        for (byte n=0; n<bracket; n++)
        {

            // if stop button pressed, cancel
            if (lastCode == BT_STOP) return false;

            // wait (to reduce camera vibrations)
            delay(waitQuarterTime);

            // trigger the camera shutter
			if (autoHalfPress && cableTrigger) cam.setHalfPress(true);
			delay(waitQuarterTime);
			if (cableTrigger) cam.setFullPress(true);
            delay(exposureTime);
			if (cableTrigger) cam.setFullPress(false);
			delay(waitQuarterTime);
			if (autoHalfPress && cableTrigger) cam.setHalfPress(false);

            // increment the shot counter
            shot++;

            // update display
            lcd.cls();
            lcd.write_P(PSTR(" STACKING    "),true);
            if (!cam.halfPressState)
            {
                lcd.writeLine_P(PSTR(" "),true);
            }
            else
            {
                lcd.writeLine_P(PSTR("3"),true,true);
            }
            lcd.write_P(PSTR("Slice "));
            lcd.write(String(i+1));
            lcd.write_P(PSTR("/"));
            lcd.writeLine(String(slices));
            lcd.write_P(PSTR("Shot "));
            lcd.write(String(shot));
            lcd.write_P(PSTR("/"));
            lcd.writeLine(String(slices*bracket));
            lcd.writeLine_P(PSTR("!\"\"\"\"\"\"\"\"\"\"\"\"#"),false,true);
            lcd.write_P(PSTR("5"),false,true);
            lcd.writeLine_P(PSTR(":Stop"));
            byte progress = ((i*14) / (slices-1));
            for (byte j=0; j<progress; j++)
            {
                if (j == 0)
                {
                    lcd.writeXY_P(j,3,PSTR("$"),false,true);
                }
                else if (j == 13)
                {
                    lcd.writeXY_P(j,3,PSTR("&"),false,true);
                }
                else
                {
                    lcd.writeXY_P(j,3,PSTR("%"),false,true);
                }
            }

            // if stop button pressed, cancel
            if (lastCode == BT_STOP) return false;

            // wait (to reduce camera vibrations)
            delay(waitQuarterTime);

        }

        // if stop button pressed, cancel
        if (lastCode == BT_STOP) return false;

    }
    return true;
}

// ----------------------------------------------------------------------------------------------------
// This function handles the stack setup
// ----------------------------------------------------------------------------------------------------

void stackMenu()
{

    boolean updateDisplay = true;
    static unsigned int slices = 80; // step count
    static unsigned int interval = 15000; // wait time (milliseconds between slices - the default of 15000 ms is to prevent flash overheating)
    static unsigned int bracket = 1; // bracket count (number of shots to take per slice)
    byte curParam = 0;
    byte lastParam = 2;
    boolean appendMode = false;
    boolean inMenu = true;

    while (inMenu)
    {
        if (updateDisplay)  	// update display
        {
            unsigned int stepDiff = abs(startPos - endPos);
            unsigned long distance = ((unsigned long)umPerRev * (unsigned long)stepDiff) / ((unsigned long)stepperMotor.stepsPerRev * 2UL);
            lcd.cls();
            lcd.write_P(PSTR(" STACK       "),true);
            if (!cam.halfPressState) lcd.writeLine_P(PSTR(" "),true);
            else lcd.writeLine_P(PSTR("3"),true,true);
            lcd.write_P(PSTR("Depth:"));
            lcd.write(String(distance));
            lcd.write_P(PSTR("'"),false,true);
            lcd.writeLine_P(PSTR("m"));
            lcd.write_P(PSTR("Slices:"),(curParam==0));
            lcd.writeLine(String(slices),(curParam==0));
            lcd.write_P(PSTR("Int:"),(curParam==1));
            lcd.write(String(interval),(curParam==1));
            lcd.writeLine_P(PSTR("ms"),(curParam==1));
            lcd.write_P(PSTR("Bracket:"),(curParam==2));
            lcd.writeLine(String(bracket),(curParam==2));
            lcd.write_P(PSTR("4"),false,true);
            lcd.writeLine_P(PSTR(":Start"));
            updateDisplay = false;
        }

        setPressState();
        byte remoteDigit = lastRemoteDigit();
        if (!pressState && (lastCode != BT_NONE))   // a button has been released
        {
            if (remoteDigit != 255) // a digit was pressed
            {
                if (curParam==0) appendMode = addDigitUI(slices,remoteDigit,appendMode);
                if (curParam==1) appendMode = addDigitUI(interval,remoteDigit,appendMode);
                if (curParam==2) appendMode = addDigitUI(bracket,remoteDigit,appendMode);
            }
            else
            {
                // enforce minimum values
                if (slices < 2) slices = 2;
                if (interval < (exposureTime  + 100)) interval = exposureTime + 100;
                if (bracket < 1) bracket = 1;

                // respond to button presses
                if (lastCode == BT_BACKLIGHT) lcd.toggleBacklight();
                else if (lastCode == BT_HALFPRESS) cam.setHalfPress(!cam.halfPressState);
                else if (lastCode == BT_FULLPRESS) cam.setFullPress(false);
                else if (lastCode == BT_UP)
                {
                    if (curParam-- == 0) curParam = lastParam; // cycle selected field up
                }
                else if (lastCode == BT_DOWN)
                {
                    if (++curParam > lastParam) curParam = 0; // cycle selected field down
                }
                else if (lastCode == BT_START) doStack(slices, interval, bracket); // do the stack
                else if (lastCode == BT_MENU) inMenu = false;
                appendMode = false;
            }
            updateDisplay = true;
        }
        else if (pressState)     // a button is being pressed
        {
            if (lastCode == BT_FULLPRESS) cam.setFullPress(true);
        }
    }
}

// ----------------------------------------------------------------------------------------------------
// This function handles the UI and stepping in the position mode
// Up/right: moves rack forward
// Down/left: moves rack backward
// Start: sets beginning position
// Stop: sets end position
// Rewind: goes to beginning position
// Fast forward: goes to end position
// ----------------------------------------------------------------------------------------------------

void positionMenu()
{

    boolean updateDisplay = true;
    boolean inMenu = true;

    while (inMenu)
    {
        if (updateDisplay) 	// update display
        {
            // calculate physical distancce between start and end points
            // Note that startPos and endPos are measured in half steps
            // This is why we multiply stepperMotor.stepsPerRev by 2, to convert it into half steps per revolution
            unsigned long stepDiff = abs(startPos - endPos);
            unsigned long distance = ((unsigned long)umPerRev * stepDiff) / ((unsigned long)stepperMotor.stepsPerRev * 2UL);

            // Show user position and depth information
            // Convert startPos and endPos to full steps to make it easy for user to perform calculations
            lcd.cls();
            lcd.write_P(PSTR(" POSITION    "),true);
            if (!cam.halfPressState) lcd.writeLine_P(PSTR(" "),true);
            else lcd.writeLine_P(PSTR("3"),true,true);
            lcd.write_P(PSTR("89"),false,true);
            lcd.write_P(PSTR(":Pos:"));
            lcd.writeLine(String(stepperMotor.pos / 2L));
            lcd.write_P(PSTR("4"),false,true);
            lcd.write_P(PSTR(":Beg:"));
            lcd.writeLine(String(startPos / 2L));
            lcd.write_P(PSTR("5"),false,true);
            lcd.write_P(PSTR(":End:"));
            lcd.writeLine(String(endPos / 2L));
            lcd.write_P(PSTR("Depth:"));
            lcd.write(String(distance));
            lcd.write_P(PSTR("'"),false,true);
            lcd.writeLine_P(PSTR("m"));
            updateDisplay = false;
        }

        setPressState();
        if (!pressState && (lastCode != BT_NONE))   // a button has been released
        {
            stepperMotor.stop();
            if (lastCode == BT_START) startPos = stepperMotor.pos;
            else if (lastCode == BT_STOP) endPos = stepperMotor.pos;
            else if ((lastCode == BT_FASTFORWARD) || (lastCode == BT_REWIND))
            {
                if (lastCode == BT_FASTFORWARD) stepperMotor.seek(endPos);
                if (lastCode == BT_REWIND) stepperMotor.seek(startPos);
                do
                {
                    setPressState();
                    stepperMotor.update();
                }
                while ((stepperMotor.isMoving) && (!pressState));
                stepperMotor.stop();
                if (relaxMotor) stepperMotor.brake();
            }
            else if (lastCode == BT_BACKLIGHT) lcd.toggleBacklight();
            else if (lastCode == BT_HALFPRESS) cam.setHalfPress(!cam.halfPressState);
            else if (lastCode == BT_FULLPRESS) cam.setFullPress(false);
            else if (lastCode == BT_MENU) inMenu = false;
            updateDisplay = true;
        }
        else if (pressState)     // a button is being pressed
        {
            if ((lastCode == BT_UP) || (lastCode == BT_RIGHT) || (lastCode == BT_DOWN) || (lastCode == BT_LEFT))
            {
                if ((lastCode == BT_UP) || (lastCode == BT_RIGHT)) stepperMotor.start(true);
                if ((lastCode == BT_DOWN) || (lastCode == BT_LEFT)) stepperMotor.start(false);
                do
                {
                    setPressState();
                    stepperMotor.update();
                }
                while ((stepperMotor.isMoving) && (pressState));
                stepperMotor.stop();
                if (relaxMotor) stepperMotor.brake();
                updateDisplay = true;
            }
            else if (lastCode == BT_FULLPRESS) cam.setFullPress(true);
        }
    }
}

// ----------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------

void motorTest()
{

    // record original settings
    unsigned int maxRPM = stepperMotor.maxRPM;
    boolean halfMode = stepperMotor.halfMode;
    boolean accelerate = stepperMotor.accelerate;
    boolean reverseMotor = stepperMotor.reverseMotor;

    // feature test
    for (byte i = 0; i<8; i++)
    {
        lcd.cls();
        lcd.writeLine_P(PSTR(" MOTOR TEST  "),true);
        lcd.write(String(maxRPM));
        lcd.writeLine_P(PSTR(" RPM"),false);
        stepperMotor.halfMode = (i & 4) > 0;
        stepperMotor.accelerate = (i & 2) > 0;
        stepperMotor.reverseMotor = (i & 1) > 0;
        if (stepperMotor.halfMode)
        {
            lcd.writeLine_P(PSTR("Half step"),false);
        }
        else
        {
            lcd.writeLine_P(PSTR("Full step"),false);
        }
        if (stepperMotor.accelerate)
        {
            lcd.writeLine_P(PSTR("Accel on"),false);
        }
        else
        {
            lcd.writeLine_P(PSTR("Accel off"),false);
        }
        if (stepperMotor.reverseMotor)
        {
            lcd.writeLine_P(PSTR("Reverse"),false);
        }
        else
        {
            lcd.writeLine_P(PSTR("Forward"),false);
        }
        stepperMotor.seek(stepperMotor.stepsPerRev * 2);
        do
        {
            stepperMotor.update();
        }
        while (stepperMotor.isMoving);
        stepperMotor.seek(0);
        do
        {
            stepperMotor.update();
        }
        while (stepperMotor.isMoving);
    }

    // timing test
    stepperMotor.halfMode = false;
    stepperMotor.accelerate = false;
    stepperMotor.reverseMotor = false;
    unsigned int revolutions = 1;
    for (byte i = 30; i<=120; i*=2)
    {
        stepperMotor.maxRPM = i;
        lcd.cls();
        lcd.writeLine_P(PSTR(" MOTOR TEST  "),true);
        lcd.write_P(PSTR("Set RPM: "),false);
        lcd.writeLine(String(i));
        unsigned long startTime = millis();
        stepperMotor.seek(stepperMotor.stepsPerRev * revolutions);
        do
        {
            stepperMotor.update();
        }
        while (stepperMotor.isMoving);
        stepperMotor.seek(0);
        do
        {
            stepperMotor.update();
        }
        while (stepperMotor.isMoving);
        unsigned long endTime = millis();
        lcd.write_P(PSTR("True RPM: "),false);
        lcd.writeLine(String( ((unsigned long)revolutions * 1000UL * 60UL) / (endTime - startTime) ));
        delay(2000);
    }

    // restore original settings
    stepperMotor.maxRPM = maxRPM;
    stepperMotor.halfMode = halfMode;
    stepperMotor.accelerate = accelerate;
    stepperMotor.reverseMotor = reverseMotor;
    if (relaxMotor) stepperMotor.brake();

}

// ----------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------

void debugMenu()
{

    boolean updateDisplay = true;
    boolean inMenu = true;

    while (inMenu)
    {
        if (updateDisplay)  	// update display
        {
            lcd.cls();
            lcd.write_P(PSTR(" DEBUG       "),true);
            if (!cam.halfPressState) lcd.writeLine_P(PSTR(" "),true);
            else lcd.writeLine_P(PSTR("3"),true,true);
            lcd.write_P(PSTR("Free Mem:"));
            lcd.writeLine(String(freeMemory()));
            lcd.writeLine_P(PSTR("1:Test Motor"));
            lcd.writeLine_P(PSTR("2:Clr EEPROM"));
            updateDisplay = false;
        }

        setPressState();
        if (!pressState && (lastCode != BT_NONE))   // a button has been released
        {
            if (lastCode == BT_1) motorTest();
            else if (lastCode == BT_2) clearEEPROM();
            else if (lastCode == BT_BACKLIGHT) lcd.toggleBacklight();
            else if (lastCode == BT_HALFPRESS) cam.setHalfPress(!cam.halfPressState);
            else if (lastCode == BT_FULLPRESS) cam.setFullPress(false);
            else if (lastCode == BT_MENU) inMenu = false;
            updateDisplay = true;
        }
        else if (pressState)     // a button is being pressed
        {
            if (lastCode == BT_FULLPRESS) cam.setFullPress(true);
        }

    }
}

// ----------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------

void loop(void)
{

    static boolean updateDisplay = true;
    if (updateDisplay)  	// update display
    {
        lcd.cls();
        lcd.write_P(PSTR(" MAIN MENU   "),true);
        if (!cam.halfPressState) lcd.writeLine_P(PSTR(" "),true);
        else lcd.writeLine_P(PSTR("3"),true,true);
        lcd.writeLine_P(PSTR("1:Position"));
        lcd.writeLine_P(PSTR("2:Stack"));
        lcd.writeLine_P(PSTR("3:Motor"));
        lcd.writeLine_P(PSTR("4:Camera"));
        lcd.writeLine_P(PSTR("5:Debug"));
        updateDisplay = false;
    }

    setPressState();
    if (!pressState && (lastCode != BT_NONE))   // a button has been released
    {
        if (lastCode == BT_1) positionMenu();
        else if (lastCode == BT_2) stackMenu();
        else if (lastCode == BT_3) motorConfig();
        else if (lastCode == BT_4) cameraConfig();
        else if (lastCode == BT_5) debugMenu();
        else if (lastCode == BT_BACKLIGHT) lcd.toggleBacklight();
        else if (lastCode == BT_HALFPRESS) cam.setHalfPress(!cam.halfPressState);
        else if (lastCode == BT_FULLPRESS) cam.setFullPress(false);
        updateDisplay = true;
    }
    else if (pressState)     // a button is being pressed
    {
        if (lastCode == BT_FULLPRESS) cam.setFullPress(true);
    }

}

// ----------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------

void setup(void)
{

    // load configuration values from EEPROM
    loadConfig();

    // initialize LCD
    // BE SURE THAT THE FOLLOWING STATEMENT IS EXECUTED VERY EARLY ON IN THE PROGRAM
    // DAMAGE TO THE LCD CAN RESULT IF CARE IS NOT TAKEN TO DO SO WITHIN 100ms OF POWER-ON
    lcd.begin(PIN_LCD_SCE, PIN_LCD_RESET, PIN_LCD_DC, PIN_LCD_SDIN, PIN_LCD_SCLK, PIN_LCD_LED);

    // initialize shift register
    shiftRegister.begin(PIN_DATA, PIN_LATCH, PIN_CLOCK);

    // enable stepper motor
    stepperMotor.enable();

    // setup pin for IR in
    pinMode(PIN_IR, INPUT);
    digitalWrite(PIN_IR, HIGH); // pulls high
    attachInterrupt(0, irFire, IR_TRIGGER);
    interrupts();

}
