#include "stepper.h"

// ----------------------------------------------------------------------------------------------------
// Stepper motor phase states - no need to modify these unless you want to use different excitation modes
// Both of the patterns make 4 full steps on the stepper motor when advanced through all 8 bits
// Advancing by one bit means advancing by one half-step
// ----------------------------------------------------------------------------------------------------

// two-phase on, full step coil states
//   A1 B1 A2 B2
// 1 1  1  0  0
// 2 1  1  0  0
// 3 0  1  1  0
// 4 0  1  1  0
// 5 0  0  1  1
// 6 0  0  1  1
// 7 1  0  0  1
// 8 1  0  0  1
// at peak current all of the time, more heat, more torque, more vibration, less precise
const byte coilA1StateFull = B11000011;
const byte coilA2StateFull = B00111100;
const byte coilB1StateFull = B11110000;
const byte coilB2StateFull = B00001111;

// half-step coil states
//   A1 B1 A2 B2
// 1 1  0  0  0
// 2 1  1  0  0
// 3 0  1  0  0
// 4 0  1  1  0
// 5 0  0  1  0
// 6 0  0  1  1
// 7 0  0  0  1
// 8 1  0  0  1
// at peak current 1/2 of the time, less heat, less torque, less vibration, more precise
const byte coilA1StateHalf = B11000001;
const byte coilA2StateHalf = B00011100;
const byte coilB1StateHalf = B01110000;
const byte coilB2StateHalf = B00000111;

stepper::stepper(shiftRegister8 *shiftRegister, byte maskEnableA, byte maskEnableB, byte maskA1, byte maskA2, byte maskB1, byte maskB2)
{

    // set member variables
    this->isMoving = false;
    this->pos = 0;
    this->goalPos = 0;
    this->reverseMotor = false;
    this->halfMode = true;
    this->accelerate = true;
    this->maxRPM = 120;
    this->curRPM = 0;
    this->accelHalfSteps = 0;
    this->stepsPerRev = 200;
    this->stepperStageBitMask = B10000000;
    this->shiftRegister = shiftRegister;
    this->maskEnableA = maskEnableA;
    this->maskEnableB = maskEnableB;
    this->maskA1 = maskA1;
    this->maskA2 = maskA2;
    this->maskB1 = maskB1;
    this->maskB2 = maskB2;

}

stepper::~stepper() {}

// ----------------------------------------------------------------------------------------------------
// This function enables the stepper motor
// ----------------------------------------------------------------------------------------------------

void stepper::enable(void)
{
    this->shiftRegister->set(this->maskEnableA | this->maskEnableB, true);
}

// ----------------------------------------------------------------------------------------------------
// This function disables the stepper motor
// ----------------------------------------------------------------------------------------------------

void stepper::disable(void)
{
    this->shiftRegister->set(this->maskEnableA | this->maskEnableB, false);
}

// ----------------------------------------------------------------------------------------------------
// This function brakes the stepper
// ----------------------------------------------------------------------------------------------------

void stepper::brake(void)
{
    this->shiftRegister->set(this->maskA1 | this->maskA2 | this->maskB1 | this->maskB2, false);
}

// ----------------------------------------------------------------------------------------------------
// This function moves the stepper by one half-step in either direction
// ----------------------------------------------------------------------------------------------------
void stepper::step(boolean forward)
{

    // move stepperStageBitMask forwards or backwards
    if ((forward && !this->reverseMotor) || (!forward && this->reverseMotor))
    {
        this->stepperStageBitMask = this->stepperStageBitMask >> 1;
        if (this->stepperStageBitMask == 0) this->stepperStageBitMask = B10000000;
    }
    else
    {
        this->stepperStageBitMask = this->stepperStageBitMask << 1;
        if (this->stepperStageBitMask == 0) this->stepperStageBitMask = B00000001;
    }

    // move the position backwards or forwards
    if (forward)
    {
        this->pos++;
    }
    else
    {
        this->pos--;
    }

    // drive the stepper
    if (this->halfMode)
    {
        this->shiftRegister->set(this->maskA1, (this->stepperStageBitMask & coilA1StateHalf) > 0, true);
        this->shiftRegister->set(this->maskA2, (this->stepperStageBitMask & coilA2StateHalf) > 0, true);
        this->shiftRegister->set(this->maskB1, (this->stepperStageBitMask & coilB1StateHalf) > 0, true);
        this->shiftRegister->set(this->maskB2, (this->stepperStageBitMask & coilB2StateHalf) > 0, true);
    }
    else
    {
        this->shiftRegister->set(this->maskA1, (this->stepperStageBitMask & coilA1StateFull) > 0, true);
        this->shiftRegister->set(this->maskA2, (this->stepperStageBitMask & coilA2StateFull) > 0, true);
        this->shiftRegister->set(this->maskB1, (this->stepperStageBitMask & coilB1StateFull) > 0, true);
        this->shiftRegister->set(this->maskB2, (this->stepperStageBitMask & coilB2StateFull) > 0, true);
    }
    this->shiftRegister->update();

}

// ----------------------------------------------------------------------------------------------------
// This function sets the destination position of the stepper
// ----------------------------------------------------------------------------------------------------
void stepper::seek(long goalPos)
{
    this->goalPos = goalPos;
}

// ----------------------------------------------------------------------------------------------------
// This function tells the stepper to continuously move
// ----------------------------------------------------------------------------------------------------

void stepper::start(boolean forward)
{
    if (forward)
    {
        this->goalPos = 2147483647;
    }
    else
    {
        this->goalPos = -2147483648;
    }
}

// ----------------------------------------------------------------------------------------------------
// This function tells the stepper to stop moving
// ----------------------------------------------------------------------------------------------------

void stepper::stop(void)
{
    this->goalPos = this->pos;
    this->isMoving = false;
    this->curRPM = 0;
    this->accelHalfSteps = 0;
}

// ----------------------------------------------------------------------------------------------------
// This function moves the stepper at the given RPM
// ----------------------------------------------------------------------------------------------------
void stepper::update(void)
{

    static unsigned long lastStepTime = 0;
    static unsigned long lastAccelTime = 0;

    // get time
    unsigned long curMicros = micros();

    // move towards goal pos
    if (this->pos != this->goalPos)
    {

        // handle acceleration and deceleration
        if (this->accelerate)
        {
            unsigned long posDiff = abs(this->goalPos - this->pos);
            if ((posDiff > this->accelHalfSteps) && (this->curRPM < this->maxRPM))   // accelerate
            {
                if ((unsigned long)(curMicros - lastAccelTime) >= (500000 / this->maxRPM))
                {
                    this->curRPM++;
                    lastAccelTime = curMicros;
                }
            }
            else if ((posDiff <= this->accelHalfSteps) && (this->curRPM > 1))     // decelerate
            {
                if ((unsigned long)(curMicros - lastAccelTime) >= (500000 / this->maxRPM))
                {
                    this->curRPM--;
                    lastAccelTime = curMicros;
                }
            }
        }
        else
        {
            this->curRPM = this->maxRPM;
        }

        // determine step interval
        unsigned long stepInterval = 60000000UL / ((unsigned long)this->curRPM * (unsigned long)this->stepsPerRev * 2UL);

        // check if it's time to take another half step
        if ((unsigned long)(curMicros - lastStepTime) >= stepInterval)
        {
            step(this->pos < goalPos);
            lastStepTime = curMicros;

            // keep track of how long it takes to accelerate so we know when to start decelerating
            if ((this->accelerate) && (this->curRPM < this->maxRPM)) this->accelHalfSteps++;
        }
        this->isMoving = true;

    }
    else
    {
        this->isMoving = false;
        this->curRPM = 0;
        this->accelHalfSteps = 0;
		lastStepTime = 0;
		lastAccelTime = 0;
    }

}
